import base64
import json

from .wallet import PrivateKey
from .client import Client


class Transaction:
    def __init__(self, client: Client, privkey: PrivateKey):
        self.client = client
        self.privkey = privkey
        self.msgs: list = []
        self.chain_id = client.get_chain_id()

    def send_tx(
        self,
        account_num: int,
        sequence: int,
        gas: int = 200000,
        fee: int = 0,
        memo: str = "",
        mode="sync",
    ) -> dict:
        base64_pubkey = base64.b64encode(bytes.fromhex(self.privkey.to_pubkey().to_hex())).decode(
            "utf-8"
        )
        tx = {
            "tx": {
                "msg": self.msgs,
                "fee": {"gas": str(gas), "amount": [{"denom": "uband", "amount": str(fee)}]},
                "memo": memo,
                "signatures": [
                    {
                        "signature": self._sign(account_num, sequence, gas, fee, memo),
                        "pub_key": {"type": "tendermint/PubKeySecp256k1", "value": base64_pubkey,},
                        "account_number": str(account_num),
                        "sequence": str(sequence),
                    }
                ],
            },
            "mode": mode,
        }
        return self.client.send_tx(tx)

    def add_request(
        self, oracle_script_id: int, calldata: bytes, ask_count: int, min_count: int, client_id: str
    ) -> None:
        msg = {
            "type": "oracle/Request",
            "value": {
                "oracle_script_id": str(oracle_script_id),
                "calldata": base64.b64encode(calldata).decode("utf-8"),
                "ask_count": str(ask_count),
                "min_count": str(min_count),
                "client_id": client_id,
                "sender": self.privkey.to_pubkey().to_address().to_acc_bech32(),
            },
        }
        self.msgs.append(msg)

    def _sign(self, account_num: int, sequence: int, gas: int, fee: int, memo: str) -> str:
        message_str = json.dumps(
            self._get_sign_message(account_num, sequence, gas, fee, memo),
            separators=(",", ":"),
            sort_keys=True,
        )
        message_bytes = message_str.encode("utf-8")

        signature = self.privkey.sign(message_bytes)

        signature_base64_str = base64.b64encode(signature).decode("utf-8")
        return signature_base64_str

    def _get_sign_message(
        self, account_num: int, sequence: int, gas: int, fee: int, memo: str
    ) -> dict:
        return {
            "chain_id": self.chain_id,
            "account_number": str(account_num),
            "fee": {"gas": str(gas), "amount": [{"amount": str(fee), "denom": "uband"}]},
            "memo": memo,
            "sequence": str(sequence),
            "msgs": self.msgs,
        }



import base64
from dataclasses import dataclass
from typing import List, Optional

from dacite import Config


DACITE_CONFIG = Config(type_hooks={int: int, bytes: base64.b64decode})


@dataclass
class DataSource(object):
    owner: str
    name: str = ""
    description: str = ""
    filename: str = ""


@dataclass
class OracleScript(object):
    owner: str
    name: str = ""
    description: str = ""
    filename: str = ""
    schema: str = ""
    source_code_url: str = ""


@dataclass
class RawRequest(object):
    data_source_id: int
    external_id: int = 0
    calldata: bytes = b""


@dataclass
class Request(object):
    oracle_script_id: int
    requested_validators: List[str]
    min_count: int
    request_height: int
    raw_requests: List[RawRequest]
    client_id: str = ""
    calldata: bytes = b""


@dataclass
class RawReport(object):
    external_id: int = 0
    data: bytes = b""


@dataclass
class Report(object):
    validator: str
    raw_reports: List[RawReport]
    in_before_resolve: bool = False


@dataclass
class RequestPacketData:
    oracle_script_id: int
    ask_count: int
    min_count: int
    client_id: str = ""
    calldata: bytes = b""


@dataclass
class ResponsePacketData(object):
    request_id: int
    request_time: int
    resolve_time: int
    resolve_status: int
    ans_count: int = 0
    client_id: str = ""
    result: bytes = b""


@dataclass
class Result(object):
    request_packet_data: RequestPacketData
    response_packet_data: ResponsePacketData


@dataclass
class RequestInfo(object):
    request: Request
    reports: List[Report]
    result: Result


@dataclass
class Account(object):
    address: str
    coins: List[dict]
    public_key: Optional[dict]
    account_number: int
    sequence: int




import requests
from dacite import from_dict

from .data import Account, DataSource, OracleScript, RequestInfo, DACITE_CONFIG


class Client(object):
    def __init__(self, rpc_url: str) -> None:
        self.rpc_url = rpc_url

    def _get(self, path, **kwargs):
        return requests.get(self.rpc_url + path, **kwargs).json()["result"]

    def send_tx(self, data: dict) -> dict:
        return requests.post(self.rpc_url + "/txs", json=data).json()

    def get_chain_id(self) -> str:
        genesis = requests.get(self.rpc_url + "/bandchain/genesis").json()
        return genesis["chain_id"]

    def get_latest_block(self) -> dict:
        return self._get("/blocks/latest")

    def get_account(self, address: str) -> Account:
        return from_dict(
            data_class=Account,
            data=self._get("/auth/accounts/{}".format(address))["value"],
            config=DACITE_CONFIG,
        )

    def get_data_source(self, id: int) -> DataSource:
        return from_dict(
            data_class=DataSource,
            data=self._get("/oracle/data_sources/{}".format(id)),
            config=DACITE_CONFIG,
        )

    def get_oracle_script(self, id: int) -> OracleScript:
        return from_dict(
            data_class=OracleScript,
            data=self._get("/oracle/oracle_scripts/{}".format(id)),
            config=DACITE_CONFIG,
        )

    def get_request_by_id(self, id: int) -> RequestInfo:
        return from_dict(
            data_class=RequestInfo,
            data=self._get("/oracle/requests/{}".format(id)),
            config=DACITE_CONFIG,
        )

    def get_latest_request(
        self, oid: int, calldata: bytes, min_count: int, ask_count: int
    ) -> RequestInfo:
        return from_dict(
            data_class=RequestInfo,
            data=self._get(
                "/oracle/request_search",
                params={
                    "oid": oid,
                    "calldata": calldata.hex(),
                    "min_count": min_count,
                    "ask_count": ask_count,
                },
            ),
            config=DACITE_CONFIG,
        )

    def get_reporters(self, validator: str) -> list:
        return self._get("/oracle/reporters/{}".format(validator))



import json
import base64

from .client import Client
from .wallet import PublicKey
from .data import Request, RequestInfo

REQUEST_DURATION = 100


class Auth:
    """
    Class for verifying that reporter information is valid to get data for report.

    :ivar client: the Client instance
    """

    def __init__(self, client: Client) -> None:
        self.client = client

    @staticmethod
    def get_msg_sign_bytes(
        chain_id: str, validator: str, request_id: str, external_id: str
    ) -> bytes:
        """
        Return bytes message using in signature verification
        """
        return str.encode(
            json.dumps(
                {
                    "chain_id": chain_id,
                    "validator": validator,
                    "request_id": request_id,
                    "external_id": external_id,
                },
                sort_keys=True,
                separators=(",", ":"),
            )
        )

    @staticmethod
    def verify_verification_message_signature(
        chain_id: str,
        validator: str,
        request_id: str,
        external_id: str,
        reporter_pubkey: str,
        signature: bytes,
    ) -> bool:
        """
        Verify the verification message using the reporter's signature
        """
        reporter = PublicKey.from_acc_bech32(reporter_pubkey)

        msg = Auth.get_msg_sign_bytes(chain_id, validator, request_id, external_id)
        return reporter.verify(msg, signature)

    def verify(
        self,
        chain_id: str,
        validator: str,
        request_id: str,
        external_id: str,
        reporter_pubkey: str,
        signature: str,
    ) -> bool:
        """
        Verify report infomation using reporter signature and on-chain request status

        :param chain_id: a chain id that request come from
        :param validator: validator address of this verify meassage
        :param request_id: a request id that request come from
        :param external_id: validator address of this verify meassage
        :param reporter_pubkey: a chain id that request come from
        :param signature: validator address of this verify meassage

        :return: True if the validator has been assigned to report data. False otherwise
        """
        if not Auth.verify_verification_message_signature(
            chain_id,
            validator,
            request_id,
            external_id,
            reporter_pubkey,
            base64.b64decode(signature),
        ):
            return False
        if not self.verify_chain_id(chain_id):
            return False

        if not self.is_reporter(validator, reporter_pubkey):
            return False

        requestInfo = self.client.get_request_by_id(request_id)

        if not self.verify_non_expired_request(requestInfo.request):
            return False
        if not self.verify_requested_validator(requestInfo.request, validator):
            return False
        if not self.verify_unsubmitted_report(requestInfo.reports, validator):
            return False

        return True

    def verify_chain_id(self, chain_id: str) -> bool:
        """
        Verify that the request comes from correct chain ID

        :param chain_id: the chain ID retrieved from the verification message

        :return: True if the chain id match with rpc client.
        """
        return self.client.get_chain_id() == chain_id

    def is_reporter(self, validator: str, reporter_pubkey: str) -> bool:
        """
        Verify that an address is a registerd reporter for validator

        :param validator: a validator on BandChain
        :param reporter_pubkey: a public key of reporter

        :return: True if the reporter is a registered reporter of the validator.
        """
        reporter = PublicKey.from_acc_bech32(reporter_pubkey).to_address().to_acc_bech32()
        reporters = self.client.get_reporters(validator)
        return reporter in reporters

    def verify_non_expired_request(self, request: Request) -> bool:
        """
        Verify the request has not expired

        :param request: a request instance

        :return: True if the request hasn't expired.
        """
        latest_block = self.client.get_latest_block()
        return (
            latest_block["block"]["header"]["height"] - request.request_height <= REQUEST_DURATION
        )

    def verify_requested_validator(self, request: Request, validator: str) -> bool:
        """
        Verify that this validator has been assigned to report this request

        :param request: a request instance
        :param validator: validator who reporter work for

        :return: True if the validator has been assigned to report data.
        """
        return validator in request.requested_validators

    def verify_unsubmitted_report(self, reports: list, validator: str) -> bool:
        """
        Verify that this validator has not reported on this request

        :param reports: list of reports on this request
        :param validator: validator who reporter work for

        :return: True if the validator hasn't reported data to chain yet
        """
        for report in reports:
            if report.validator == validator:
                return False
        return True

